        .thumb
        .syntax unified

        .global stacktop, _start
        
        .section .text
        
        @ vector table
        .word stacktop
        .word reset_handler+1
        .word nmi_handler+1
        .word hardfault_handler+1
        .word memmanage_handler+1
        .word busfault_handler+1
        .word usagefault_handler+1
        .word except7_handler+1
        .word except8_handler+1
        .word except9_handler+1
        .word except10_handler+1
        .word svcall_handler+1
        .word except12_handler+1
        .word except13_handler+1
        .word pendsv_handler+1
        .word systick_handler+1
        .word irq0_handler+1
        .word irq1_handler+1
        .word irq2_handler+1
        .word irq3_handler+1
        .word irq4_handler+1
        .word irq5_handler+1
        .word irq6_handler+1
        .word irq7_handler+1

reset_handler:
        nop
        b _start
        
nmi_handler:
        nop
        b nmi_handler
        
hardfault_handler:
        nop
        b hardfault_handler
        
memmanage_handler:
        nop
        b memmanage_handler
        
busfault_handler:
        nop
        b busfault_handler
        
usagefault_handler:
        nop
        b usagefault_handler
        
except7_handler:
        nop
        b except7_handler
        
except8_handler:
        nop
        b except8_handler
        
except9_handler:
        nop
        b except9_handler
        
except10_handler:
        nop
        b except10_handler
        
svcall_handler:
        nop
        b svcall_handler
        
except12_handler:
        nop
        b except12_handler
        
except13_handler:
        nop
        b except13_handler
        
pendsv_handler:
        nop
        b pendsv_handler
        
systick_handler:
        nop
        b systick_handler
        
irq0_handler:
        nop
        b irq0_handler
        
irq1_handler:
        nop
        b irq1_handler
        
irq2_handler:
        nop
        b irq2_handler
        
irq3_handler:
        nop
        b irq3_handler
        
irq4_handler:
        nop
        b irq4_handler
        
irq5_handler:
        nop
        b irq5_handler
        
irq6_handler:
        nop
        b irq6_handler
        
irq7_handler:
        nop
        b irq7_handler

_start:        
        bl configureclock   @ switch to 120MHz
        bl configurepins    @ configure external bus pins
        bl configureemc     @ configure static CS3

        @ enable LED output
        @ configure P9_0 as GPIO P4.12
        ldr r0, =0x40086480
        ldr r1, =(1<<5)         @ enable high-speed driver
        str r1, [r0]
        @ configure GPIO 4.12 as output to control LED
        ldr r0, =0x400F6010
        ldr r1, =(1<<12)
        str r1, [r0]
        
1:
        @ do read on external bus using CS3
        ldr r1, =0x1f000000 @ base address for CS3
        ldrb r0,[r1]
        
        @ approx 100ms delay when running at 120MHz
        ldr r3, =3000000        
2:      subs r3, #1
        bne  2b

        @ toggle LED
        @ point to P4 toggle register
        ldr  r0, =0x400F6310
        @ bit 12
        ldr r1, =(1<<12)
        @ toggle P4.12
        str r1,[r0]
        
        b    1b


@ enable and configure EMC        
configureemc:
        @ load registers with values from table
        ldr  r0,=emc_init_table
        ldr  r3,=emc_init_table_end
1:      ldr  r1,[r0]    @ register address
        ldr  r2,[r0,#4] @ init value
        str  r2,[r1]
        add  r0,#8
        cmp  r0,r3
        bne  1b
        @ done
        bx   lr

@ register addresses        
.set EMC_CONTROL,     0x40005000
.set STATICCONFIG3,   0x40005260
.set STATICWAITWEN3,  0x40005264  @ CS -> WR delay
.set STATICWAITOEN3,  0x40005268  @ CS/ADDR -> OE delay
.set STATICWAITRD3,   0x4000526c  @ CS->RD delay
.set STATICWAITPAGE3, 0x40005270  @
.set STATICWAITWR3,   0x40005274  @ CS->WR delay
.set STATICWAITTURN3, 0x40005278  @
.set CLK_EMCDIV_CFG,  0x40051478

@ bit positions in EMC_CONTROL register        
.set EMC_CONTROL_ENABLE,  0
.set EMC_CONTROL_MIRROR,  1
        
@ bit positions in static config register
.set CNF_MEM_WIDTH,        0    @ bus width 0==8, 1==16, 2==32
.set CNF_PAGE_MODE,        3    @ enalbe page mode (len=4)
.set CNF_CS_HIGH,          6    @ CS polarity 0=low, 1==high
.set CNF_BLS_READ,         7    @ byte lane active for 0=wr, 1=rd/wr
.set CNF_EXT_WAIT,         8    @ enable extended wait
.set CNF_BUFF_EN,         19    @ buffer enable
.set CNF_WR_PROT,         20    @ enable write protect
       
@ bit positions in EMC_CONTROL register
.set  EMC_CONTROL_ENABLE,    0
.set  EMC_CONTROL_MIRROR,    1
.set  EMC_CONTROL_LOWPWR,    2

@ value for EMC_CONTROL        
.set EMC_CONTROL_value, 0
.set EMC_CONTROL_value, EMC_CONTROL_value | (1 << EMC_CONTROL_ENABLE)
.set EMC_CONTROL_value, EMC_CONTROL_value | (0 << EMC_CONTROL_MIRROR)
.set EMC_CONTROL_value, EMC_CONTROL_value | (0 << EMC_CONTROL_LOWPWR)

@ value for EMCDIV         
.set  CLK_EMCDIV_CFG_value, 0
.set  CLK_EMCDIV_CFG_value, CLK_EMCDIV_CFG_value | (1 << 0)  @ RUN
.set  CLK_EMCDIV_CFG_value, CLK_EMCDIV_CFG_value | (0 << 1)  @ AUTO
.set  CLK_EMCDIV_CFG_value, CLK_EMCDIV_CFG_value | (0 << 2)  @ WAKEUP
.set  CLK_EMCDIV_CFG_value, CLK_EMCDIV_CFG_value | (0 << 1)  @ DIV
        
@ value to be put into STATICCONFIG3        
.set STATICCONFIG3_value, 0
.set STATICCONFIG3_value,  STATICCONFIG3_value | (0 << CNF_MEM_WIDTH) @ bus width 0==8, 1==16, 2==32
.set STATICCONFIG3_value,  STATICCONFIG3_value | (0 << CNF_PAGE_MODE) @ enable page mode (len=4)
.set STATICCONFIG3_value,  STATICCONFIG3_value | (0 << CNF_CS_HIGH)   @ CS polarity 0=low, 1==high
.set STATICCONFIG3_value,  STATICCONFIG3_value | (1 << CNF_BLS_READ)  @ byte lane active for 0=wr, 1=rd/wr
.set STATICCONFIG3_value,  STATICCONFIG3_value | (0 << CNF_EXT_WAIT)  @ enable extended wait
.set STATICCONFIG3_value,  STATICCONFIG3_value | (0 << CNF_BUFF_EN)   @ buffer enable
.set STATICCONFIG3_value,  STATICCONFIG3_value | (0 << CNF_WR_PROT)   @ enable write protect

        
emc_init_table:
        @ Table of initialization values for EMC controller
        .word EMC_CONTROL,        EMC_CONTROL_value
        .word CLK_EMCDIV_CFG,     CLK_EMCDIV_CFG_value
        .word STATICCONFIG3,      STATICCONFIG3_value
        .word STATICWAITOEN3,     0    @ 15 max   CS->OE delay: N*8.333ns  
        .word STATICWAITRD3,      5    @ 31 max   CS length:  2*(N+1)*8.333ns 
        .word STATICWAITTURN3,    2    @ 15 max   bus turn around after read
        .word STATICWAITWEN3,     0    @ 15 max   CS->WE delay: (N+1)*8.333ns
        .word STATICWAITWR3,      5    @ 31 max   WE length:    (N+1)*8.333ns
emc_init_table_end:     
        

@ Set up to run core from PLL1 at 120MHz
configureclock: 
        @ register addresses
        .equ SFSCLK3_CTRL,      0x40086c0c
        .equ XTAL_OSC_CTRL,     0x40050018
        .equ BASE_M3_CLK_CTRL,  0x4005006c
        .equ BASE_OUT_CLK_CTRL, 0x400500ac
        .equ PLL1_STAT,         0x40050040
        .equ PLL1_CTRL,         0x40050044

        @ PLL1_CTRL register bit field shift values
        .equ PLL1_CTRL_PD,         0
        .equ PLL1_CTRL_BYPASS,     1
        .equ PLL1_CTRL_FBSEL,      6
        .equ PLL1_CTRL_DIRECT,     7
        .equ PLL1_CTRL_PSEL,       8
        .equ PLL1_CTRL_AUTOBLOCK, 11
        .equ PLL1_CTRL_NSEL,      12
        .equ PLL1_CTRL_MSEL,      16
        .equ PLL1_CTRL_CLK_SEL,   24

@@@ Assume worst case: xtal off, PLL1 running at low freq
        
        @ Enable xtal oscillator (12MHz)

        ldr r0,=XTAL_OSC_CTRL
        ldr r1,=0
        str r1,[r0]

        @ wait at least 100us.  Running at 200MHz, this loop should
        @ still take 100us.  It will be even longer if we're running
	@ at 12MHz or 96MHz, but not so long as to be a problem.

        ldr   r1,=200000
1:      subs  r1,#1
        bne   1b

        @ select xtal as source for BASE_M3_CLK

        ldr r0,=BASE_M3_CLK_CTRL
        ldr r1,=(0x06<<24)              @ xtal source
        str r1,[r0]


        @ switching from 12MHz to 120MHz requires an intermediate
	@ stopover at 96MHz

        @ connect PLL1 output to CLKOUT node.

        ldr r0,=BASE_OUT_CLK_CTRL
        ldr r1,=(0x01<<24)   @ IRC
        str r1,[r0]

        @ connect CLKOUT node to CLK3 pin

        ldr r0,=SFSCLK3_CTRL
        ldr r1,=(1<<0)|(1<<4)  @ source==1 disable pullup
        str r1,[r0]

        
        @ configure PLL1 for N=1, M=8, post-divide=2 to get 96MHz from 12MHz
        
        .set PLL1_CTRL_VALUE, 0
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (0 << PLL1_CTRL_PD)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (0 << PLL1_CTRL_BYPASS)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (1 << PLL1_CTRL_FBSEL)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (0 << PLL1_CTRL_DIRECT)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (0 << PLL1_CTRL_PSEL)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (0 << PLL1_CTRL_AUTOBLOCK)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (0 << PLL1_CTRL_NSEL)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (7 << PLL1_CTRL_MSEL)
        @.set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (6 << PLL1_CTRL_CLK_SEL)  @ xtal
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE | (1 << PLL1_CTRL_CLK_SEL)  @ IRC
        
        ldr r0,=PLL1_CTRL
        ldr r1,=PLL1_CTRL_VALUE
        str r1,[r0]

        @ wait for PLL1 lock
        
        nop
        nop
        ldr  r0,=PLL1_STAT
1:      ldr  r1,[r0]
        ands r1,#1
        beq  1b

        @ select PLL1 as source for M3 clock

        ldr r0,=BASE_M3_CLK_CTRL
        ldr r1,=(0x09<<24)              @ PLL1
        str r1,[r0]

        @ change PLL1 multipler from X8 to X10 to jump from 96MHz to 120MHz
        
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE & ~(0xff << PLL1_CTRL_MSEL)
        .set PLL1_CTRL_VALUE, PLL1_CTRL_VALUE |  (9 << PLL1_CTRL_MSEL)
        
        ldr r0,=PLL1_CTRL
        ldr r1,=PLL1_CTRL_VALUE
        str r1,[r0]

        @ wait for PLL1 lock
        
        nop
        nop
        ldr  r0,=PLL1_STAT
1:      ldr  r1,[r0]
        ands r1,#1
        beq  1b

        @ Now were running at 120MHz
        bx   lr


@ set up pin assignments from table
configurepins:  
        ldr  r0,=pintable
        ldr  r3,=pintableend
1:      ldr  r1,[r0]       @ SFS<pin> register address
        ldr  r2,[r0,#4]    @ mode value
        orr  r2,#(1<<5)    @ enable high slew rate
        orr  r2,#(1<<6)    @ enable input buffer
        orr  r2,#(1<<7)    @ disable glitch filter
        str  r2,[r1]
        add  r0,#8
        cmp  r0,r3
        bne  1b
        bx   lr


@ Pin assignment table for external memories

pintable:
            @   SFSreg   mode         node     EMC
            @ ---------- ----         ----    -----

        .word 0x4008632c,  3        @ P6_11   CKEOUT
        
        .word 0x40086094,  3        @ P1_5    /CS0
        .word 0x4008630c,  3        @ P6_3    /CS1
        .word 0x400866b0,  2        @ PD_12   /CS2  
        .word 0x400866ac,  2        @ PD_11   /CS3
        
        .word 0x40086090,  3        @ P1_4    /BLS0 
        .word 0x40086318,  1        @ P6_6    /BLS1 
        .word 0x4008608c,  3        @ P1_3      /OE
        .word 0x40086098,  3        @ P1_6      /WE
        
        .word 0x4008609c,  3        @ P1_7       D0
        .word 0x400860a0,  3        @ P1_8       D1
        .word 0x400860a4,  3        @ P1_9       D2
        .word 0x400860a8,  3        @ P1_10      D3
        .word 0x400860ac,  3        @ P1_11      D4
        .word 0x400860b0,  3        @ P1_12      D5
        .word 0x400860b4,  3        @ P1_13      D6
        .word 0x400860b8,  3        @ P1_14      D7
        .word 0x40086290,  2        @ P5_4       D8
        .word 0x40086294,  2        @ P5_5       D9
        .word 0x40086298,  2        @ P5_6      D10
        .word 0x4008629c,  2        @ P5_7      D11
        .word 0x40086280,  2        @ P5_0      D12
        .word 0x40086284,  2        @ P5_1      D13
        .word 0x40086288,  2        @ P5_2      D14
        .word 0x4008628c,  2        @ P5_3      D15
        .word 0x40086688,  2        @ PD_2      D16
        .word 0x4008668c,  2        @ PD_3      D17
        .word 0x40086690,  2        @ PD_4      D18
        .word 0x40086694,  2        @ PD_5      D19
        .word 0x40086698,  2        @ PD_6      D20
        .word 0x4008669c,  2        @ PD_7      D21
        .word 0x400866a0,  2        @ PD_8      D22
        .word 0x400866a4,  2        @ PD_9      D23
        .word 0x40086714,  3        @ PE_5      D24
        .word 0x40086718,  3        @ PE_6      D25
        .word 0x4008671c,  3        @ PE_7      D26
        .word 0x40086720,  3        @ PE_8      D27
        .word 0x40086724,  3        @ PE_9      D28
        .word 0x40086728,  3        @ PE_10     D29
        .word 0x4008672c,  3        @ PE_11     D30
        .word 0x40086730,  3        @ PE_12     D31
        
        .word 0x40086124,  3        @ P2_9       A0
        .word 0x40086128,  3        @ P2_10      A1
        .word 0x4008612c,  3        @ P2_11      A2
        .word 0x40086130,  3        @ P2_12      A3
        .word 0x40086134,  3        @ P2_13      A4
        .word 0x40086080,  2        @ P1_0       A5
        .word 0x40086084,  2        @ P1_1       A6
        .word 0x40086088,  2        @ P1_2       A7
        .word 0x40086120,  3        @ P2_8       A8
        .word 0x4008611c,  3        @ P2_7       A9
        .word 0x40086118,  2        @ P2_6      A10
        .word 0x40086108,  2        @ P2_2      A11
        .word 0x40086104,  2        @ P2_1      A12
        .word 0x40086100,  2        @ P2_0      A13
        .word 0x40086320,  1        @ P6_8      A14
        .word 0x4008631c,  1        @ P6_7      A15
        .word 0x400866c0,  2        @ PD_16     A16
        .word 0x400866bc,  2        @ PD_15     A17
        .word 0x40086700,  3        @ PE_0      A18
        .word 0x40086704,  3        @ PE_1      A19
        .word 0x40086708,  3        @ PE_2      A20
        .word 0x4008670c,  3        @ PE_3      A21
        .word 0x40086710,  3        @ PE_4      A22
        .word 0x40086510,  3        @ PA_4      A23

pintableend:
        